#include "lbs_ii.h"

Define_Module( LBS_II )

void LBS_II::activity()
{
    // get parameters
    simtime_t delta_t = par("delta_t");

    // search servers and save reference to its workload parameter
    // as well as their rate
    cModule *compound = parentModule();

    for (int i = 0; i<serverCount; i++) {
        cModule *svr = compound->submodule("server", i);
        if ((svr==0) || (strcmp(svr->className(), "Server")!=0)) {
            throw cException("server[] contains non-Server");
        }
        server[i] = (Server*) svr;
        rate[i] = svr->par("rate");
    }

    // create update-message and issue it immediately
    cMessage *updateMsg = new cMessage("update-balancer");
    scheduleAt(simTime() + delta_t, updateMsg);
    update();

    // message loop
    for (;;) {
        // get message
        cMessage *msg = receive();

        if (msg==updateMsg) {
            // update received, so do it and issue the message again delta_t secs later
            update();
            scheduleAt(simTime() + delta_t, updateMsg);
        } else {
            // send the message to a server
            send(msg, "out", chooseServer());
        }
    }

}

void LBS_II::update()
{
    // search minimum workload
    int i;
    double load = 100000000;
    for (i = 0; i<serverCount; i++) {
        // get queue length
        long l = server[i]->getJobsInSystem();
        if (l>0) --l;
        // calculate performance measure and save it
        double bl = (double) l/rate[i];
        balancing[i] = bl;
        // record smallest workload
        if (bl<load) {
            load = bl;
        }
    }

    // calculate weights
    if (load==0.0) {
        // minimum workload is zero => use different set of equations
        for (i = 0; i<serverCount; i++) {
            if (balancing[i]!=0.0) balancing[i] = 1/(1+balancing[i]);
            else balancing[i] = 1.0;
        }
    } else {
        // standard weight calculation
        for (i = 0; i<serverCount; i++) {
            if (balancing[i]!=0.0) balancing[i] = load/balancing[i];
            else balancing[i] = 1.0;
        }
    }

    // total sum of weights
    double wSum = 0.0;
    for (i = 0; i<serverCount; i++) wSum += balancing[i];
    
    // normalize weights
    for (i = 0; i<serverCount; i++) {
        balancing[i] /= wSum;
        ev << "balancing[" << i << "] = " << balancing[i] << endl;
    }

    // change balancing into distribution
    for (i = 1; i<serverCount; i++) balancing[i] += balancing[i-1];
}

int LBS_II::chooseServer()
{
    // get random value
    double rVal = uniform(0, 1, rg);

    // get server no via inverse transformation method
    int i = 0;
    while ((i<serverCount) && (balancing[i]<rVal)) ++i;
    if (i==serverCount) throw cException("error while choosing server!");

    ev << "chose server " << i << endl;

    return i;
}

void LBS_II::initialize()
{
    // get parameters
    rg = par("random_generator");
    serverCount = par("server_count");
    // create arrays
    server = new Server *[serverCount];
    rate = new double[serverCount];
    balancing = new double[serverCount];
}

void LBS_II::finish()
{
    delete [] server;
    delete [] rate;
    delete [] balancing;
}
